/*
 * Device driver for the INI-9XXXU/UW or INIC-940/950  PCI SCSI Controller.
 *
 * Written for 386bsd and FreeBSD by
 *	Winston Hung		<winstonh@initio.com>
 *
 * Copyright (c) 1997-1999 Initio Corp
 * Copyright (c) 2000-2002 Ken Westerback
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR OR HIS RELATIVES BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF MIND, USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Ported from OpenBSD by Stefan Farfeleder <stefan@fafoe.dyndns.org>
 *	$OpenBSD: iha.c,v 1.16 2002/11/14 02:31:46 krw Exp $
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>

#include <vm/vm.h>

#include <cam/cam.h>
#include <cam/cam_ccb.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt_sim.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>

#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include <dev/iha/iha.h>

/*
 * SCSI Rate Table, indexed by FLAG_SCSI_RATE field of tcs_flags.
 */
static const uint8_t iha_rate_tbl[8] = {
	/* fast 20		  */
	/* nanosecond divide by 4 */
	12,	/* 50ns,  20M	  */
	18,	/* 75ns,  13.3M	  */
	25,	/* 100ns, 10M	  */
	31,	/* 125ns, 8M	  */
	37,	/* 150ns, 6.6M	  */
	43,	/* 175ns, 5.7M	  */
	50,	/* 200ns, 5M	  */
	62	/* 250ns, 4M	  */
};
#define	IHA_MAX_PERIOD	62

static void	iha_abort_ccb(struct iha_softc *, union ccb *, uint8_t);
static void	iha_action(struct cam_sim *sim, union ccb *ccb);
static int	iha_alloc_scbs(struct iha_softc *);
static void	iha_append_done_scb(struct iha_softc *, struct iha_scb *,
		    uint8_t);
static void	iha_append_free_scb(struct iha_softc *, struct iha_scb *);
static void	iha_append_pend_scb(struct iha_softc *, struct iha_scb *);
static int	iha_attach(device_t);
static void	iha_bad_seq(struct iha_softc *);
static void	iha_busfree(struct iha_softc *);
static int	iha_cam_attach(struct iha_softc *);
static void	iha_cam_release(struct iha_softc *);
static const char *iha_chipname(device_t);
static uint8_t	iha_data_over_run(struct iha_scb *);
static void	iha_del_pend_scb(struct iha_softc *, struct iha_scb *);
static int	iha_detach(device_t);
static void	iha_disable_ints(struct iha_softc *);
static void	iha_enable_ints(struct iha_softc *);
static bus_dmamap_callback_t iha_dmamap_helper;
static bus_dmamap_callback_t iha_exec_scb;
static void	iha_done_scb(struct iha_softc *, struct iha_scb *);
static int	iha_init_tulip(struct iha_softc *);
static void	iha_intr(void *);
static void	iha_main(struct iha_softc *);
static void	iha_mark_busy_scb(struct iha_scb *);
static int	iha_msgin(struct iha_softc *);
static int	iha_msgin_extended(struct iha_softc *);
static int	iha_msgin_ignore_wid_resid(struct iha_softc *);
static int	iha_msgin_sdtr(struct iha_softc *);
static int	iha_msgout(struct iha_softc *, uint8_t);
static void	iha_msgout_abort(struct iha_softc *, uint8_t);
static int	iha_msgout_extended(struct iha_softc *);
static int	iha_msgout_reject(struct iha_softc *);
static int	iha_msgout_sdtr(struct iha_softc *);
static int	iha_msgout_wdtr(struct iha_softc *);
static int	iha_next_state(struct iha_softc *);
static void	iha_poll(struct cam_sim *sim);
static int	iha_probe(device_t);
static void	iha_push_pend_scb(struct iha_softc *, struct iha_scb *);
static void	iha_push_sense_request(struct iha_softc *, struct iha_scb *);
static __inline uint8_t iha_read_8(struct iha_softc *, bus_size_t);
static __inline uint32_t iha_read_32(struct iha_softc *, bus_size_t);
static void	iha_read_eeprom(struct iha_softc *, struct iha_nvram *);
static void	iha_release(struct iha_softc *);
static int	iha_resel(struct iha_softc *);
static void	iha_reset_chip(struct iha_softc *);
static void	iha_reset_dma(struct iha_softc *);
static void	iha_reset_scsi_bus(struct iha_softc *);
static void	iha_reset_tcs(struct tcs *, uint8_t);
static void	iha_scsi(struct iha_softc *);
static void	iha_scsi_io(struct iha_softc *, union ccb *);
static struct	iha_scb *iha_find_pend_scb(struct iha_softc *);
static struct	iha_scb *iha_pop_done_scb(struct iha_softc *);
static struct	iha_scb *iha_pop_free_scb(struct iha_softc *);
static void	iha_se2_instr(struct iha_softc *, uint8_t);
static uint16_t	iha_se2_rd(struct iha_softc *, uint8_t);
static void	iha_select(struct iha_softc *, struct iha_scb *, uint8_t);
static void	iha_set_ssig(struct iha_softc *, uint8_t, uint8_t);
static void	iha_start_scb(struct iha_softc *, struct iha_scb *);
static int	iha_state_1(struct iha_softc *);
static int	iha_state_2(struct iha_softc *);
static int	iha_state_3(struct iha_softc *);
static int	iha_state_4(struct iha_softc *);
static int	iha_state_5(struct iha_softc *);
static int	iha_state_6(struct iha_softc *);
static int	iha_state_8(struct iha_softc *);
static int	iha_status_msg(struct iha_softc *);
static void	iha_sync_done(struct iha_softc *);
static timeout_t iha_timeout;
static int	iha_wait(struct iha_softc *, uint8_t);
static void	iha_wide_done(struct iha_softc *);
static void	iha_write_8(struct iha_softc *, bus_size_t, uint8_t);
static void	iha_write_32(struct iha_softc *, bus_size_t, uint32_t);
static int	iha_xfer_data(struct iha_softc *, struct iha_scb *, int);
static int	iha_xpad_in(struct iha_softc *);
static int	iha_xpad_out(struct iha_softc *);


static device_method_t iha_methods[] = {
	DEVMETHOD(device_probe,		iha_probe),
	DEVMETHOD(device_attach,	iha_attach),
	DEVMETHOD(device_detach,	iha_detach),
	{ 0, 0 }
};

static driver_t iha_driver = {
	"iha",
	iha_methods,
	sizeof(struct iha_softc)
};

static devclass_t	iha_devclass;
DRIVER_MODULE(iha, pci, iha_driver, iha_devclass, 0, 0);


#define	PCI_VENDOR_INITIO		0x1101
#define	PCI_PRODUCT_INITIO_INIC940	0x9400
#define	PCI_PRODUCT_INITIO_INIC941	0x9401
#define	PCI_PRODUCT_INITIO_INIC950	0x9500

#if defined(IHA_DEBUG)
#define	IHA_TRACE	printf("iha %s:%d\n", __func__, __LINE__)
#define	IHA_PRINTF(x)	printf x
#else
#define	IHA_TRACE	/* nothing */
#define	IHA_PRINTF(x)	/* nothing */
#endif

static __inline uint8_t
iha_read_8(struct iha_softc *sc, bus_size_t offset)
{
	return (bus_space_read_1(sc->sc_iot, sc->sc_ioh, offset));
}

static __inline uint32_t
iha_read_32(struct iha_softc *sc, bus_size_t offset)
{
	return (bus_space_read_4(sc->sc_iot, sc->sc_ioh, offset));
}

static __inline void
iha_write_8(struct iha_softc *sc, bus_size_t offset, uint8_t data)
{
	bus_space_write_1(sc->sc_iot, sc->sc_ioh, offset, data);
}

static __inline void
iha_write_32(struct iha_softc *sc, bus_size_t offset, uint32_t data)
{
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, offset, data);
}

/*
 * iha_intr - the interrupt service routine for the iha driver
 */
static void
iha_intr(void *arg)
{
	struct iha_softc *sc;
	int s;

	IHA_TRACE;

	sc = arg;

	if ((iha_read_8(sc, TUL_STAT0) & INTPD) == 0)
		return;

	s = splbio(); /* XXX - Or are interrupts off when ISR's are called? */

	if (sc->sc_semaph != SEMAPH_IN_MAIN) {
		/* XXX - need these inside a splbio()/splx()? */
		iha_disable_ints(sc);
		sc->sc_semaph = SEMAPH_IN_MAIN;

		iha_main(sc);

		sc->sc_semaph = ~SEMAPH_IN_MAIN;
		iha_enable_ints(sc);
	}

	splx(s);
}

/*
 * iha_init_tulip - initialize the inic-940/950 card and the rest of the
 *		    iha_softc structure supplied
 */
static int
iha_init_tulip(struct iha_softc *sc)
{
	struct iha_scb *scb;
	struct iha_nvram_channel *nvc;
	struct iha_nvram iha_nvram;
	int i, error;

	iha_read_eeprom(sc, &iha_nvram);

	nvc = &iha_nvram.nvm_channel[0];

	sc->sc_target_max     = (nvc->nvc_targets == 16) ? 16 : 8;
	sc->sc_adapter_target = nvc->nvc_id;

	/*
	 * fill in the rest of the iha_softc fields
	 */
	sc->sc_semaph	 = ~SEMAPH_IN_MAIN;
	sc->sc_jsstatus0 = 0;
	sc->sc_actscb	 = NULL;

	TAILQ_INIT(&sc->sc_freescb);
	TAILQ_INIT(&sc->sc_pendscb);
	TAILQ_INIT(&sc->sc_donescb);

	error = iha_alloc_scbs(sc);
	if (error != 0)
		return (error);

	for (i = 0, scb = sc->sc_scb; i < IHA_MAX_SCB; i++, scb++) {
		scb->scb_tag_id = i;
		scb->scb_sg_paddr = sc->sc_scb_phys
		    + i * sizeof(struct iha_scb)
		    + offsetof(struct iha_scb, scb_sg_list);

		scb->scb_sense_len   = sizeof(struct scsi_sense_data);
		scb->scb_sense_paddr = sc->sc_scb_phys
		    + i * sizeof(struct iha_scb)
		    + offsetof(struct iha_scb, scb_scsi_sense_data);

		error = bus_dmamap_create(sc->sc_buf_dmat, 0, &scb->scb_dmamap);
		if (error != 0) {
			device_printf(sc->sc_dev, "couldn't create SCB DMA "
			    "map, error = %d\n", error);
			return (error);
		}
		TAILQ_INSERT_TAIL(&sc->sc_freescb, scb, scb_next);
	}

	/* Mask all the interrupts */
	iha_disable_ints(sc);

	/* Stop any I/O and reset the scsi module */
	iha_reset_dma(sc);
	iha_write_8(sc, TUL_SCTRL0, RSMOD);

	/* Program HBA's SCSI ID */
	iha_write_8(sc, TUL_SID, sc->sc_adapter_target << 4);

	/*
	 * Configure the channel as requested by the NVRAM settings read
	 * into iha_nvram by iha_read_eeprom() above.
	 */

	if ((nvc->nvc_cfg & CFG_EN_PAR) != 0)
		sc->sc_sconf1 = (SCONFIG0DEFAULT | SPCHK);
	else
		sc->sc_sconf1 = (SCONFIG0DEFAULT);
	iha_write_8(sc, TUL_SCONFIG0, sc->sc_sconf1);

	/* selection time out in units of 1.6385 millisecond = 250 ms */
	iha_write_8(sc, TUL_STIMO, 153);

	/* Enable desired SCSI termination configuration read from eeprom */
	iha_write_8(sc, TUL_DCTRL0,
	    nvc->nvc_cfg & (CFG_ACT_TERM1 | CFG_ACT_TERM2));

	iha_write_8(sc, TUL_GCTRL1, ((nvc->nvc_cfg & CFG_AUTO_TERM) >> 4)
	    | (iha_read_8(sc, TUL_GCTRL1) & (~ATDEN)));

	for (i = 0; i < IHA_MAX_TARGETS; i++) {
		sc->sc_tcs[i].tcs_flags = nvc->nvc_target_flags[i];
		iha_reset_tcs(&sc->sc_tcs[i], sc->sc_sconf1);
	}

	iha_reset_chip(sc);

	return (0);
}

/*
 * iha_reset_dma - abort any active DMA xfer, reset tulip FIFO.
 */
static void
iha_reset_dma(struct iha_softc *sc)
{
	if ((iha_read_8(sc, TUL_ISTUS1) & XPEND) != 0) {
		/* if DMA xfer is pending, abort DMA xfer */
		iha_write_8(sc, TUL_DCMD, ABTXFR);
		/* wait Abort DMA xfer done */
		while ((iha_read_8(sc, TUL_ISTUS0) & DABT) == 0)
			;
	}

	iha_write_8(sc, TUL_SCTRL0, RSFIFO);
}

/*
 * iha_pop_free_scb - return the first free SCB, or NULL if there are none.
 */
static struct iha_scb *
iha_pop_free_scb(struct iha_softc *sc)
{
	struct iha_scb *scb;
	int s;

	s = splbio();

	scb = TAILQ_FIRST(&sc->sc_freescb);

	if (scb != NULL) {
		scb->scb_status = STATUS_RENT;
		TAILQ_REMOVE(&sc->sc_freescb, scb, scb_next);
	}

	splx(s);

	return (scb);
}

/*
 * iha_append_free_scb - append the supplied SCB to the tail of the
 *                       sc_freescb queue after clearing and resetting
 *			 everything possible.
 */
static void
iha_append_free_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	int s;

	s = splbio();

	if (scb == sc->sc_actscb)
		sc->sc_actscb = NULL;

	scb->scb_status  = STATUS_QUEUED;
	scb->scb_ha_stat = HOST_OK;
	scb->scb_ta_stat = SCSI_STATUS_OK;

	scb->scb_next_stat = 0;
	scb->scb_sg_idx    = 0;
	scb->scb_sg_max    = 0;
	scb->scb_flags     = 0;
	scb->scb_target    = 0;
	scb->scb_lun       = 0;
	scb->scb_buf_len   = 0;
	scb->scb_sg_len    = 0;
	scb->scb_cdb_len   = 0;
	scb->scb_ident     = 0;
	scb->scb_tag_msg   = 0;
	scb->scb_buf_paddr = 0;

	scb->scb_ccb = NULL;
	scb->scb_tcs = NULL;

	bzero( scb->scb_cdb,             sizeof(scb->scb_cdb));
	bzero(&scb->scb_scsi_sense_data, sizeof(scb->scb_scsi_sense_data));
	bzero( scb->scb_sg_list,         sizeof(scb->scb_sg_list));

	/*
	 * scb_tag_id, scb_sg_paddr, scb_sense_len, scb_sg_list and
	 * scb_sense_paddr are set at initialization and never change.
	 */

	TAILQ_INSERT_TAIL(&sc->sc_freescb, scb, scb_next);

	splx(s);
}

static void
iha_append_pend_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	/* ASSUMPTION: only called within a splbio()/splx() pair */

	if (scb == sc->sc_actscb)
		sc->sc_actscb = NULL;

	scb->scb_status = STATUS_QUEUED;

	TAILQ_INSERT_TAIL(&sc->sc_pendscb, scb, scb_next);
}

static void
iha_push_pend_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	int s;

	s = splbio();

	if (scb == sc->sc_actscb)
		sc->sc_actscb = NULL;

	scb->scb_status = STATUS_QUEUED;

	TAILQ_INSERT_HEAD(&sc->sc_pendscb, scb, scb_next);

	splx(s);
}

/*
 * iha_find_pend_scb - scan the pending queue for a SCB that can be
 *		       processed immediately. Return NULL if none found
 *		       and a pointer to the SCB if one is found. If there
 *		       is an active SCB, return NULL!
 */
static struct iha_scb *
iha_find_pend_scb(struct iha_softc *sc)
{
	struct iha_scb *scb;
	struct tcs *tcs;
	int s;

	s = splbio();

	if (sc->sc_actscb != NULL)
		scb = NULL;

	else
		TAILQ_FOREACH(scb, &sc->sc_pendscb, scb_next) {
			tcs = scb->scb_tcs;

			if ((scb->scb_tag_msg) != 0) {
				/*
				 * A Tagged I/O. OK to start If no
				 * non-tagged I/O is active on the same
				 * target
				 */
				if (tcs->tcs_nontag_scb == NULL)
					break;

			} else	if (scb->scb_cdb[0] == REQUEST_SENSE) {
				/*
				 * OK to do a non-tagged request sense
				 * even if a non-tagged I/O has been
				 * started, because we don't allow any
				 * disconnect during a request sense op
				 */
				break;

			} else	if (tcs->tcs_tag_cnt == 0) {
				/*
				 * No tagged I/O active on this target,
				 * ok to start a non-tagged one if one
				 * is not already active
				 */
				if (tcs->tcs_nontag_scb == NULL)
					break;
			}
		}

	splx(s);

	return (scb);
}

/*
 * iha_del_pend_scb - remove scb from sc_pendscb
 */
static void
iha_del_pend_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	int s;

	s = splbio();

	TAILQ_REMOVE(&sc->sc_pendscb, scb, scb_next);

	splx(s);
}

static void
iha_mark_busy_scb(struct iha_scb *scb)
{
	int  s;

	s = splbio();

	scb->scb_status = STATUS_BUSY;

	if (scb->scb_tag_msg == 0)
		scb->scb_tcs->tcs_nontag_scb = scb;
	else
		scb->scb_tcs->tcs_tag_cnt++;

	splx(s);
}

static void
iha_append_done_scb(struct iha_softc *sc, struct iha_scb *scb, uint8_t ha_stat)
{
	struct tcs *tcs;
	int s;

	s = splbio();

	if (scb->scb_ccb != NULL)
		untimeout(iha_timeout, scb, scb->scb_ccb->ccb_h.timeout_ch);

	if (scb == sc->sc_actscb)
		sc->sc_actscb = NULL;

	tcs = scb->scb_tcs;

	if (scb->scb_tag_msg != 0) {
		if (tcs->tcs_tag_cnt)
			tcs->tcs_tag_cnt--;
	} else if (tcs->tcs_nontag_scb == scb)
		tcs->tcs_nontag_scb = NULL;

	scb->scb_status = STATUS_QUEUED;
	scb->scb_ha_stat = ha_stat;

	TAILQ_INSERT_TAIL(&sc->sc_donescb, scb, scb_next);

	splx(s);
}

static struct iha_scb *
iha_pop_done_scb(struct iha_softc *sc)
{
	struct iha_scb *scb;
	int s;

	s = splbio();

	scb = TAILQ_FIRST(&sc->sc_donescb);

	if (scb != NULL) {
		scb->scb_status = STATUS_RENT;
		TAILQ_REMOVE(&sc->sc_donescb, scb, scb_next);
	}

	splx(s);

	return (scb);
}

/*
 * iha_abort_ccb - find the SCB associated with the supplied ccb and
 *                 stop all processing on it, moving it to the done
 *                 queue with the supplied host status value.
 */
static void
iha_abort_ccb(struct iha_softc *sc, union ccb *ccb, uint8_t ha_stat)
{
	struct iha_scb *scb;
	int i, s;

	s = splbio();

	/* Check the pending queue for the SCB pointing to ccb */

	TAILQ_FOREACH(scb, &sc->sc_pendscb, scb_next)
		if (scb->scb_ccb == ccb) {
			iha_del_pend_scb(sc, scb);
			iha_append_done_scb(sc, scb, ha_stat);
			splx(s);
			return;
		}

	/*
	 * If that didn't work, check all BUSY/SELECTING SCB's for one
	 * pointing to ccb.
	 */

	for (i = 0, scb = sc->sc_scb; i < IHA_MAX_SCB; i++, scb++)
		switch (scb->scb_status) {
		case STATUS_BUSY:
		case STATUS_SELECT:
			if (scb->scb_ccb == ccb) {
				iha_append_done_scb(sc, scb, ha_stat);
				splx(s);
				return;
			}
			break;
		default:
			break;
		}

	splx(s);
}

/*
 * iha_bad_seq - a SCSI bus phase was encountered out of the
 *               correct/expected sequence. Reset the SCSI bus.
 */
static void
iha_bad_seq(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;

	if (scb != NULL)
		iha_append_done_scb(sc, scb, HOST_BAD_PHAS);

	iha_reset_scsi_bus(sc);
	iha_reset_chip(sc);
}

/*
 * iha_push_sense_request - obtain auto sense data by pushing the
 *                          SCB needing it back onto the pending
 *			    queue with a REQUEST_SENSE CDB.
 */
static void
iha_push_sense_request(struct iha_softc *sc, struct iha_scb *scb)
{
	IHA_TRACE;

	scb->scb_buf_len   = scb->scb_sense_len;
	scb->scb_buf_paddr = scb->scb_sense_paddr;

	scb->scb_flags &= ~(CAM_SCATTER_VALID | CAM_DIR_MASK);
	scb->scb_flags |= CAM_SEND_SENSE | CAM_DIR_IN;

	scb->scb_ident &= ~MSG_IDENTIFY_DISCFLAG;

	scb->scb_tag_msg = 0;
	scb->scb_ta_stat = SCSI_STATUS_OK;

	bzero(scb->scb_cdb, sizeof(scb->scb_cdb));

	scb->scb_cdb_len = 6;
	scb->scb_cdb[0]  = REQUEST_SENSE;
	scb->scb_cdb[4]  = scb->scb_sense_len;

	iha_push_pend_scb(sc, scb);
}

/*
 * iha_main - process the active SCB, taking one off pending and making it
 *            active if necessary, and any done SCB's created as
 *            a result until there are no interrupts pending and no pending
 *            SCB's that can be started.
 */
static void
iha_main(struct iha_softc *sc)
{
	struct iha_scb *scb;

	IHA_TRACE;

	for (;;) {
iha_scsi_label:
		iha_scsi(sc);

		while ((scb = iha_pop_done_scb(sc)) != NULL) {

			/* XXX: merge with iha_done (see NetBSD) */
			switch (scb->scb_ta_stat) {
			case SCSI_STATUS_CMD_TERMINATED:
			case SCSI_STATUS_ACA_ACTIVE:
			case SCSI_STATUS_CHECK_COND:
				scb->scb_tcs->tcs_flags &=
				    ~(FLAG_SYNC_DONE | FLAG_WIDE_DONE);

				if ((scb->scb_flags & CAM_SEND_SENSE) != 0)
					/* Check condition on check condition*/
					scb->scb_ha_stat = HOST_BAD_PHAS;

				else {
					iha_push_sense_request(sc, scb);
					goto iha_scsi_label;
				}
				break;

			default:
				if ((scb->scb_flags & CAM_SEND_SENSE) != 0)
					/*
					 * Return the original
					 * SCSI_STATUS_CHECK_COND, not the
					 * status of the request sense
					 * command!
					 */
					scb->scb_ta_stat =
						SCSI_STATUS_CHECK_COND;
				break;
			}

			iha_done_scb(sc, scb);
		}

		/*
		 * If there are no interrupts pending, or we can't start
		 * a pending sc, break out of the for(;;). Otherwise
		 * continue the good work with another call to
		 * iha_scsi().
		 */
		if (((iha_read_8(sc, TUL_STAT0) & INTPD) == 0)
		    && (iha_find_pend_scb(sc) == NULL))
			break;
	}
}

/*
 * iha_scsi - service any outstanding interrupts. If there are none, try to
 *            start another SCB currently in the pending queue.
 */
static void
iha_scsi(struct iha_softc *sc)
{
	struct iha_scb *scb;
	struct tcs *tcs;
	uint8_t stat;

	/* service pending interrupts asap */

	stat = iha_read_8(sc, TUL_STAT0);
	if ((stat & INTPD) != 0) {
		sc->sc_jsstatus0 = stat;
		sc->sc_jsstatus1 = iha_read_8(sc, TUL_STAT1);
		sc->sc_jsint     = iha_read_8(sc, TUL_SISTAT);

		sc->sc_phase = sc->sc_jsstatus0 & PH_MASK;

		if ((sc->sc_jsint & SRSTD) != 0) {
			iha_reset_scsi_bus(sc);
			return;
		}

		if ((sc->sc_jsint & RSELED) != 0) {
			iha_resel(sc);
			return;
		}

		if ((sc->sc_jsint & (STIMEO | DISCD)) != 0) {
			iha_busfree(sc);
			return;
		}

		if ((sc->sc_jsint & (SCMDN | SBSRV)) != 0) {
			iha_next_state(sc);
			return;
		}

		if ((sc->sc_jsint & SELED) != 0)
			iha_set_ssig(sc, 0, 0);
	}

	/*
	 * There were no interrupts pending which required action elsewhere, so
	 * see if it is possible to start the selection phase on a pending SCB
	 */
	if ((scb = iha_find_pend_scb(sc)) == NULL)
		return;

	tcs = scb->scb_tcs;

	/* program HBA's SCSI ID & target SCSI ID */
	iha_write_8(sc, TUL_SID,
	    (sc->sc_adapter_target << 4) | scb->scb_target);

	iha_write_8(sc, TUL_SYNCM, tcs->tcs_js_period);

	if ((tcs->tcs_flags & FLAG_NO_NEG_WIDE) == 0 ||
	    (tcs->tcs_flags & FLAG_NO_NEG_SYNC) == 0)
		iha_select(sc, scb, SELATNSTOP);

	else if (scb->scb_tag_msg != 0)
		iha_select(sc, scb, SEL_ATN3);

	else
		iha_select(sc, scb, SEL_ATN);

}

/*
 * iha_data_over_run - return HOST_OK for all SCSI opcodes where BufLen
 *                     is an 'Allocation Length'. All other SCSI opcodes
 *		       get HOST_DO_DU as they SHOULD have xferred all the
 *		       data requested.
 *
 *		       The list of opcodes using 'Allocation Length' was
 * 		       found by scanning all the SCSI-3 T10 drafts. See
 *		       www.t10.org for the curious with a .pdf reader.
 */
static uint8_t
iha_data_over_run(struct iha_scb *scb)
{
	switch (scb->scb_cdb[0]) {
	case 0x03: /* Request Sense                   SPC-2 */
	case 0x12: /* Inquiry                         SPC-2 */
	case 0x1a: /* Mode Sense (6 byte version)     SPC-2 */
	case 0x1c: /* Receive Diagnostic Results      SPC-2 */
	case 0x23: /* Read Format Capacities          MMC-2 */
	case 0x29: /* Read Generation                 SBC   */
	case 0x34: /* Read Position                   SSC-2 */
	case 0x37: /* Read Defect Data                SBC   */
	case 0x3c: /* Read Buffer                     SPC-2 */
	case 0x42: /* Read Sub Channel                MMC-2 */
	case 0x43: /* Read TOC/PMA/ATIP               MMC   */

	/* XXX - 2 with same opcode of 0x44? */
	case 0x44: /* Read Header/Read Density Suprt  MMC/SSC*/

	case 0x46: /* Get Configuration               MMC-2 */
	case 0x4a: /* Get Event/Status Notification   MMC-2 */
	case 0x4d: /* Log Sense                       SPC-2 */
	case 0x51: /* Read Disc Information           MMC   */
	case 0x52: /* Read Track Information          MMC   */
	case 0x59: /* Read Master CUE                 MMC   */
	case 0x5a: /* Mode Sense (10 byte version)    SPC-2 */
	case 0x5c: /* Read Buffer Capacity            MMC   */
	case 0x5e: /* Persistant Reserve In           SPC-2 */
	case 0x84: /* Receive Copy Results            SPC-2 */
	case 0xa0: /* Report LUNs                     SPC-2 */
	case 0xa3: /* Various Report requests         SBC-2/SCC-2*/
	case 0xa4: /* Report Key                      MMC-2 */
	case 0xad: /* Read DVD Structure              MMC-2 */
	case 0xb4: /* Read Element Status (Attached)  SMC   */
	case 0xb5: /* Request Volume Element Address  SMC   */
	case 0xb7: /* Read Defect Data (12 byte ver.) SBC   */
	case 0xb8: /* Read Element Status (Independ.) SMC   */
	case 0xba: /* Report Redundancy               SCC-2 */
	case 0xbd: /* Mechanism Status                MMC   */
	case 0xbe: /* Report Basic Redundancy         SCC-2 */

		return (HOST_OK);
		break;

	default:
		return (HOST_DO_DU);
		break;
	}
}

/*
 * iha_next_state - prcess the current SCB as requested in it's
 *                  scb_next_stat member.
 */
static int
iha_next_state(struct iha_softc *sc)
{
	if (sc->sc_actscb == NULL)
		return (-1);

	switch (sc->sc_actscb->scb_next_stat) {
	case 1:
		if (iha_state_1(sc) == 3)
			goto state_3;
		break;

	case 2:
		switch (iha_state_2(sc)) {
		case 3:
			goto state_3;
		case 4:
			goto state_4;
		default:
			break;
		}
		break;

	case 3:
	state_3:
		if (iha_state_3(sc) == 4)
			goto state_4;
		break;

	case 4:
	state_4:
		switch (iha_state_4(sc)) {
		case 0:
			return (0);
		case 6:
			goto state_6;
		default:
			break;
		}
		break;

	case 5:
		switch (iha_state_5(sc)) {
		case 4:
			goto state_4;
		case 6:
			goto state_6;
		default:
			break;
		}
		break;

	case 6:
	state_6:
		iha_state_6(sc);
		break;

	case 8:
		iha_state_8(sc);
		break;

	default:
#ifdef IHA_DEBUG_STATE
		printf("[debug] -unknown state: %i-\n",
		    sc->sc_actscb->scb_next_stat);
#endif
		iha_bad_seq(sc);
		break;
	}

	return (-1);
}

/*
 * iha_state_1 - selection is complete after a SELATNSTOP. If the target
 *               has put the bus into MSG_OUT phase start wide/sync
 *               negotiation. Otherwise clear the FIFO and go to state 3,
 *	    	 which will send the SCSI CDB to the target.
 */
static int
iha_state_1(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;
	struct tcs *tcs;
	uint16_t flags;

	iha_mark_busy_scb(scb);

	tcs = scb->scb_tcs;

	iha_write_8(sc, TUL_SCONFIG0, tcs->tcs_sconfig0);

	/*
	 * If we are in PHASE_MSG_OUT, send
	 *     a) IDENT message (with tags if appropriate)
	 *     b) WDTR if the target is configured to negotiate wide xfers
	 *     ** OR **
	 *     c) SDTR if the target is configured to negotiate sync xfers
	 *	  but not wide ones
	 *
	 * If we are NOT, then the target is not asking for anything but
	 * the data/command, so go straight to state 3.
	 */
	if (sc->sc_phase == PHASE_MSG_OUT) {
		iha_write_8(sc, TUL_SCTRL1, (ESBUSIN | EHRSL));
		iha_write_8(sc, TUL_SFIFO, scb->scb_ident);

		if (scb->scb_tag_msg != 0) {
			iha_write_8(sc, TUL_SFIFO, scb->scb_tag_msg);
			iha_write_8(sc, TUL_SFIFO, scb->scb_tag_id);
		}

		flags = tcs->tcs_flags;
		if ((flags & FLAG_NO_NEG_WIDE) == 0) {
			if (iha_msgout_wdtr(sc) == -1)
				return (-1);
		} else if ((flags & FLAG_NO_NEG_SYNC) == 0) {
			if (iha_msgout_sdtr(sc) == -1)
				return (-1);
		}

	} else {
		iha_write_8(sc, TUL_SCTRL0, RSFIFO);
		iha_set_ssig(sc, REQ | BSY | SEL | ATN, 0);
	}

	return (3);
}

/*
 * iha_state_2 - selection is complete after a SEL_ATN or SEL_ATN3. If the SCSI
 *		 CDB has already been send, go to state 4 to start the data
 *               xfer. Otherwise reset the FIFO and go to state 3, sending
 *		 the SCSI CDB.
 */
static int
iha_state_2(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;

	iha_mark_busy_scb(scb);

	iha_write_8(sc, TUL_SCONFIG0, scb->scb_tcs->tcs_sconfig0);

	if ((sc->sc_jsstatus1 & CPDNE) != 0)
		return (4);

	iha_write_8(sc, TUL_SCTRL0, RSFIFO);

	iha_set_ssig(sc, REQ | BSY | SEL | ATN, 0);

	return (3);
}

/*
 * iha_state_3 - send the SCSI CDB to the target, processing any status
 *		 or other messages received until that is done or
 *               abandoned.
 */
static int
iha_state_3(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;
	uint16_t flags;

	for (;;)
		switch (sc->sc_phase) {
		case PHASE_CMD_OUT:
			bus_space_write_multi_1(sc->sc_iot, sc->sc_ioh,
			    TUL_SFIFO, scb->scb_cdb, scb->scb_cdb_len);
			if (iha_wait(sc, XF_FIFO_OUT) == -1)
				return (-1);
			else if (sc->sc_phase == PHASE_CMD_OUT) {
				iha_bad_seq(sc);
				return (-1);
			} else
				return (4);

		case PHASE_MSG_IN:
			scb->scb_next_stat = 3;
			if (iha_msgin(sc) == -1)
				return (-1);
			break;

		case PHASE_STATUS_IN:
			if (iha_status_msg(sc) == -1)
				return (-1);
			break;

		case PHASE_MSG_OUT:
			flags = scb->scb_tcs->tcs_flags;
			if ((flags & FLAG_NO_NEG_SYNC) != 0) {
				if (iha_msgout(sc, MSG_NOOP) == -1)
					return (-1);
			} else if (iha_msgout_sdtr(sc) == -1)
				return (-1);
			break;

		default:
			printf("[debug] -s3- bad phase = %d\n", sc->sc_phase);
			iha_bad_seq(sc);
			return (-1);
		}
}

/*
 * iha_state_4 - start a data xfer. Handle any bus state
 *               transitions until PHASE_DATA_IN/_OUT
 *               or the attempt is abandoned. If there is
 *               no data to xfer, go to state 6 and finish
 *               processing the current SCB.
 */
static int
iha_state_4(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;

	if ((scb->scb_flags & CAM_DIR_MASK) == CAM_DIR_NONE)
		return (6); /* Both dir flags set => NO xfer was requested */

	for (;;) {
		if (scb->scb_buf_len == 0)
			return (6);

		switch (sc->sc_phase) {
		case PHASE_STATUS_IN:
			if ((scb->scb_flags & CAM_DIR_MASK) != CAM_DIR_NONE)
				scb->scb_ha_stat = iha_data_over_run(scb);
			if (iha_status_msg(sc) == -1)
				return (-1);
			break;

		case PHASE_MSG_IN:
			scb->scb_next_stat = 4;
			if (iha_msgin(sc) == -1)
				return (-1);
			break;

		case PHASE_MSG_OUT:
			if ((sc->sc_jsstatus0 & SPERR) != 0) {
				scb->scb_buf_len = 0;
				scb->scb_ha_stat = HOST_SPERR;
				if (iha_msgout(sc, MSG_INITIATOR_DET_ERR) == -1)
					return (-1);
				else
					return (6);
			} else {
				if (iha_msgout(sc, MSG_NOOP) == -1)
					return (-1);
			}
			break;

		case PHASE_DATA_IN:
			return (iha_xfer_data(sc, scb, CAM_DIR_IN));

		case PHASE_DATA_OUT:
			return (iha_xfer_data(sc, scb, CAM_DIR_OUT));

		default:
			iha_bad_seq(sc);
			return (-1);
		}
	}
}

/*
 * iha_state_5 - handle the partial or final completion of the current
 *               data xfer. If DMA is still active stop it. If there is
 *		 more data to xfer, go to state 4 and start the xfer.
 *               If not go to state 6 and finish the SCB.
 */
static int
iha_state_5(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;
	struct iha_sg_element *pSg;
	uint32_t cnt;
	uint16_t period;
	uint8_t stat;
	long xcnt;  /* cannot use unsigned!! see code: if (xcnt < 0) */
	int i;

	cnt = iha_read_32(sc, TUL_STCNT0) & TCNT;

	/*
	 * Stop any pending DMA activity and check for parity error.
	 */

	if ((iha_read_8(sc, TUL_DCMD) & XDIR) != 0) {
		/* Input Operation */
		if ((sc->sc_jsstatus0 & SPERR) != 0)
			scb->scb_ha_stat = HOST_SPERR;

		if ((iha_read_8(sc, TUL_ISTUS1) & XPEND) != 0) {
			iha_write_8(sc, TUL_DCTRL0,
			    iha_read_8(sc, TUL_DCTRL0) | SXSTP);
			while (iha_read_8(sc, TUL_ISTUS1) & XPEND)
				;
		}

	} else {
		/* Output Operation */
		if ((sc->sc_jsstatus1 & SXCMP) == 0) {
			period = scb->scb_tcs->tcs_js_period;
			if ((period & PERIOD_WIDE_SCSI) != 0)
				cnt += (iha_read_8(sc, TUL_SFIFOCNT) &
				    FIFOC) << 1;
			else
				cnt += (iha_read_8(sc, TUL_SFIFOCNT) &
				    FIFOC);
		}

		if ((iha_read_8(sc, TUL_ISTUS1) & XPEND) != 0) {
			iha_write_8(sc, TUL_DCMD, ABTXFR);
			do
				stat = iha_read_8(sc, TUL_ISTUS0);
			while ((stat & DABT) == 0);
		}

		if ((cnt == 1) && (sc->sc_phase == PHASE_DATA_OUT)) {
			if (iha_wait(sc, XF_FIFO_OUT) == -1)
				return (-1);
			cnt = 0;

		} else if ((sc->sc_jsstatus1 & SXCMP) == 0)
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);
	}

	if (cnt == 0) {
		scb->scb_buf_len = 0;
		return (6);
	}

	/* Update active data pointer and restart the I/O at the new point */

	xcnt		 = scb->scb_buf_len - cnt; /* xcnt == bytes xferred */
	scb->scb_buf_len = cnt;			   /* cnt  == bytes left    */

	if ((scb->scb_flags & CAM_SCATTER_VALID) != 0) {
		pSg = &scb->scb_sg_list[scb->scb_sg_idx];
		for (i = scb->scb_sg_idx; i < scb->scb_sg_max; pSg++, i++) {
			xcnt -= pSg->sg_len;
			if (xcnt < 0) {
				/* XXX: NetBSD uses le32toh()/htole32() here */
				xcnt += pSg->sg_len;

				pSg->sg_ptr += xcnt;
				pSg->sg_len -= xcnt;

				/* XXX: NetBSD has a bus_dmamap_sync() here */
				scb->scb_buf_paddr += (i - scb->scb_sg_idx)
					* sizeof(struct iha_sg_element);
				scb->scb_sg_len	= scb->scb_sg_max - i;
				scb->scb_sg_idx	= i;

				return (4);
			}
		}
		return (6);

	} else
		scb->scb_buf_paddr += xcnt;

	return (4);
}

/*
 * iha_state_6 - finish off the active scb (may require several
 *               iterations if PHASE_MSG_IN) and return -1 to indicate
 *		 the bus is free.
 */
static int
iha_state_6(struct iha_softc *sc)
{
	for (;;)
		switch (sc->sc_phase) {
		case PHASE_STATUS_IN:
			if (iha_status_msg(sc) == -1)
				return (-1);
			break;

		case PHASE_MSG_IN:
			sc->sc_actscb->scb_next_stat = 6;
			if (iha_msgin(sc) == -1)
				return (-1);
			break;

		case PHASE_MSG_OUT:
			if (iha_msgout(sc, MSG_NOOP) == -1)
				return (-1);
			break;

		case PHASE_DATA_IN:
			if (iha_xpad_in(sc) == -1)
				return (-1);
			break;

		case PHASE_DATA_OUT:
			if (iha_xpad_out(sc) == -1)
				return (-1);
			break;

		default:
			iha_bad_seq(sc);
			return (-1);
		}
}

/*
 * iha_state_8 - reset the active device and all busy SCBs using it
 */
static int
iha_state_8(struct iha_softc *sc)
{
	struct iha_scb *scb;
	uint32_t i;
	uint8_t tar;

	if (sc->sc_phase == PHASE_MSG_OUT) {
		iha_write_8(sc, TUL_SFIFO, MSG_BUS_DEV_RESET);

		scb = sc->sc_actscb;

		/* This SCB finished correctly -- resetting the device */
		iha_append_done_scb(sc, scb, HOST_OK);

		iha_reset_tcs(scb->scb_tcs, sc->sc_sconf1);

		tar = scb->scb_target;
		for (i = 0, scb = sc->sc_scb; i < IHA_MAX_SCB; i++, scb++)
			if (scb->scb_target == tar)
				switch (scb->scb_status) {
				case STATUS_BUSY:
					iha_append_done_scb(sc, scb,
					    HOST_DEV_RST);
					break;

				case STATUS_SELECT:
					iha_push_pend_scb(sc, scb);
					break;

				default:
					break;
				}

		sc->sc_flags |= FLAG_EXPECT_DISC;

		if (iha_wait(sc, XF_FIFO_OUT) == -1)
			return (-1);
	}

	iha_bad_seq(sc);
	return (-1);
}

/*
 * iha_xfer_data - initiate the DMA xfer of the data
 */
static int
iha_xfer_data(struct iha_softc *sc, struct iha_scb *scb, int direction)
{
	uint32_t xferlen;
	uint8_t xfertype;

	if ((scb->scb_flags & CAM_DIR_MASK) != direction)
		return (6); /* wrong direction, abandon I/O */

	iha_write_32(sc, TUL_STCNT0, scb->scb_buf_len);

	if ((scb->scb_flags & CAM_SCATTER_VALID) == 0) {
		xferlen  = scb->scb_buf_len;
		xfertype = (direction == CAM_DIR_IN) ? ST_X_IN : ST_X_OUT;

	} else {
		xferlen  = scb->scb_sg_len * sizeof(struct iha_sg_element);
		xfertype = (direction == CAM_DIR_IN) ? ST_SG_IN : ST_SG_OUT;
	}

	iha_write_32(sc, TUL_DXC,  xferlen);
	iha_write_32(sc, TUL_DXPA, scb->scb_buf_paddr);
	iha_write_8 (sc, TUL_DCMD, xfertype);

	iha_write_8(sc, TUL_SCMD,
	    (direction == CAM_DIR_IN) ? XF_DMA_IN : XF_DMA_OUT);

	scb->scb_next_stat = 5;

	return (0);
}

static int
iha_xpad_in(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;

	if ((scb->scb_flags & CAM_DIR_MASK) != CAM_DIR_NONE)
		scb->scb_ha_stat = HOST_DO_DU;

	for (;;) {
		if ((scb->scb_tcs->tcs_js_period & PERIOD_WIDE_SCSI) != 0)
			iha_write_32(sc, TUL_STCNT0, 2);
		else
			iha_write_32(sc, TUL_STCNT0, 1);

		switch (iha_wait(sc, XF_FIFO_IN)) {
		case -1:
			return (-1);

		case PHASE_DATA_IN:
			iha_read_8(sc, TUL_SFIFO);
			break;

		default:
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);
			return (6);
		}
	}
}

static int
iha_xpad_out(struct iha_softc *sc)
{
	struct iha_scb *scb = sc->sc_actscb;

	if ((scb->scb_flags & CAM_DIR_MASK) != CAM_DIR_NONE)
		scb->scb_ha_stat = HOST_DO_DU;

	for (;;) {
		if ((scb->scb_tcs->tcs_js_period & PERIOD_WIDE_SCSI) != 0)
			iha_write_32(sc, TUL_STCNT0, 2);
		else
			iha_write_32(sc, TUL_STCNT0, 1);

		iha_write_8(sc, TUL_SFIFO, 0);

		switch (iha_wait(sc, XF_FIFO_OUT)) {
		case -1:
			return (-1);

		case PHASE_DATA_OUT:
			break;

		default:
			/* Disable wide CPU to allow read 16 bits */
			iha_write_8(sc, TUL_SCTRL1, EHRSL);
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);
			return (6);
		}
	}
}

static int
iha_status_msg(struct iha_softc *sc)
{
	struct iha_scb *scb;
	uint8_t msg;
	int phase;

	if ((phase = iha_wait(sc, CMD_COMP)) == -1)
		return (-1);

	scb = sc->sc_actscb;

	scb->scb_ta_stat = iha_read_8(sc, TUL_SFIFO);

	if (phase == PHASE_MSG_OUT) {
		if ((sc->sc_jsstatus0 & SPERR) == 0)
			iha_write_8(sc, TUL_SFIFO, MSG_NOOP);
		else
			iha_write_8(sc, TUL_SFIFO, MSG_PARITY_ERROR);

		return (iha_wait(sc, XF_FIFO_OUT));

	} else if (phase == PHASE_MSG_IN) {
		msg = iha_read_8(sc, TUL_SFIFO);

		if ((sc->sc_jsstatus0 & SPERR) != 0)
			switch (iha_wait(sc, MSG_ACCEPT)) {
			case -1:
				return (-1);
			case PHASE_MSG_OUT:
				iha_write_8(sc, TUL_SFIFO, MSG_PARITY_ERROR);
				return (iha_wait(sc, XF_FIFO_OUT));
			default:
				iha_bad_seq(sc);
				return (-1);
			}

		if (msg == MSG_CMDCOMPLETE) {
			if ((scb->scb_ta_stat & (SCSI_STATUS_INTERMED |
			    SCSI_STATUS_BUSY)) == SCSI_STATUS_INTERMED) {
				iha_bad_seq(sc);
				return (-1);
			}
			sc->sc_flags |= FLAG_EXPECT_DONE_DISC;
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);
			return (iha_wait(sc, MSG_ACCEPT));
		}

		if ((msg == MSG_LINK_CMD_COMPLETE)
		    || (msg == MSG_LINK_CMD_COMPLETEF)) {
			if ((scb->scb_ta_stat & (SCSI_STATUS_INTERMED |
			    SCSI_STATUS_BUSY)) == SCSI_STATUS_INTERMED)
				return (iha_wait(sc, MSG_ACCEPT));
		}
	}

	iha_bad_seq(sc);
	return (-1);
}

/*
 * iha_busfree - SCSI bus free detected as a result of a TIMEOUT or
 *		 DISCONNECT interrupt. Reset the tulip FIFO and
 *		 SCONFIG0 and enable hardware reselect. Move any active
 *		 SCB to sc_donescb list. Return an appropriate host status
 *		 if an I/O was active.
 */
static void
iha_busfree(struct iha_softc *sc)
{
	struct iha_scb *scb;

	iha_write_8(sc, TUL_SCTRL0,   RSFIFO);
	iha_write_8(sc, TUL_SCONFIG0, SCONFIG0DEFAULT);
	iha_write_8(sc, TUL_SCTRL1,   EHRSL);

	scb = sc->sc_actscb;

	if (scb != NULL) {
		if (scb->scb_status == STATUS_SELECT)
			/* selection timeout   */
			iha_append_done_scb(sc, scb, HOST_SEL_TOUT);
		else
			/* Unexpected bus free */
			iha_append_done_scb(sc, scb, HOST_BAD_PHAS);
	}
}

static void
iha_reset_scsi_bus(struct iha_softc *sc)
{
	struct iha_scb *scb;
	struct tcs *tcs;
	int i, s;

	s = splbio();

	iha_reset_dma(sc);

	for (i = 0, scb = sc->sc_scb; i < IHA_MAX_SCB; i++, scb++)
		switch (scb->scb_status) {
		case STATUS_BUSY:
			iha_append_done_scb(sc, scb, HOST_SCSI_RST);
			break;

		case STATUS_SELECT:
			iha_push_pend_scb(sc, scb);
			break;

		default:
			break;
		}

	for (i = 0, tcs = sc->sc_tcs; i < IHA_MAX_TARGETS; i++, tcs++)
		iha_reset_tcs(tcs, sc->sc_sconf1);

	splx(s);
}

/*
 * iha_resel - handle a detected SCSI bus reselection request.
 */
static int
iha_resel(struct iha_softc *sc)
{
	struct iha_scb *scb;
	struct tcs *tcs;
	uint8_t tag, target, lun, msg, abortmsg;

	if (sc->sc_actscb != NULL) {
		if (sc->sc_actscb->scb_status == STATUS_SELECT)
			iha_push_pend_scb(sc, sc->sc_actscb);
		sc->sc_actscb = NULL;
	}

	target = iha_read_8(sc, TUL_SBID);
	lun    = iha_read_8(sc, TUL_SALVC) & MSG_IDENTIFY_LUNMASK;

	tcs = &sc->sc_tcs[target];

	iha_write_8(sc, TUL_SCONFIG0, tcs->tcs_sconfig0);
	iha_write_8(sc, TUL_SYNCM,    tcs->tcs_js_period);

	abortmsg = MSG_ABORT; /* until a valid tag has been obtained */

	if (tcs->tcs_nontag_scb != NULL)
		/* There is a non-tagged I/O active on the target */
		scb = tcs->tcs_nontag_scb;

	else {
		/*
		 * Since there is no active non-tagged operation
		 * read the tag type, the tag itself, and find
		 * the appropriate scb by indexing sc_scb with
		 * the tag.
		 */

		switch (iha_wait(sc, MSG_ACCEPT)) {
		case -1:
			return (-1);
		case PHASE_MSG_IN:
			iha_write_32(sc, TUL_STCNT0, 1);
			if (iha_wait(sc, XF_FIFO_IN) == -1)
				return (-1);
			break;
		default:
			goto abort;
		}

		msg = iha_read_8(sc, TUL_SFIFO);	/* Read Tag Msg */

		if ((msg < MSG_SIMPLE_Q_TAG) || (msg > MSG_ORDERED_Q_TAG))
			goto abort;

		switch (iha_wait(sc, MSG_ACCEPT)) {
		case -1:
			return (-1);
		case PHASE_MSG_IN:
			iha_write_32(sc, TUL_STCNT0, 1);
			if (iha_wait(sc, XF_FIFO_IN) == -1)
				return (-1);
			break;
		default:
			goto abort;
		}

		tag = iha_read_8(sc, TUL_SFIFO);	/* Read Tag ID */
		scb = &sc->sc_scb[tag];

		abortmsg = MSG_ABORT_TAG; /* Now that we have valdid tag! */
	}

	if (scb->scb_target != target ||
	    scb->scb_lun    != lun    ||
	    scb->scb_status != STATUS_BUSY) {
abort:
		iha_msgout_abort(sc, abortmsg);
		return (-1);
	}

	sc->sc_actscb = scb;

	if (iha_wait(sc, MSG_ACCEPT) == -1)
		return (-1);

	return (iha_next_state(sc));
}

static int
iha_msgin(struct iha_softc *sc)
{
	uint16_t flags;
	uint8_t msg;
	int phase;

	for (;;) {
		if ((iha_read_8(sc, TUL_SFIFOCNT) & FIFOC) > 0)
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);

		iha_write_32(sc, TUL_STCNT0, 1);

		phase = iha_wait(sc, XF_FIFO_IN);
		msg   = iha_read_8(sc, TUL_SFIFO);

		switch (msg) {
		case MSG_DISCONNECT:
			sc->sc_flags |= FLAG_EXPECT_DISC;
			if (iha_wait(sc, MSG_ACCEPT) != -1)
				iha_bad_seq(sc);
			phase = -1;
			break;
		case MSG_SAVEDATAPOINTER:
		case MSG_RESTOREPOINTERS:
		case MSG_NOOP:
			phase = iha_wait(sc, MSG_ACCEPT);
			break;
		case MSG_MESSAGE_REJECT:
			/* XXX - need to clear FIFO like other 'Clear ATN'? */
			iha_set_ssig(sc, REQ | BSY | SEL | ATN, 0);
			flags = sc->sc_actscb->scb_tcs->tcs_flags;
			if ((flags & FLAG_NO_NEG_SYNC) == 0)
				iha_set_ssig(sc, REQ | BSY | SEL, ATN);
			phase = iha_wait(sc, MSG_ACCEPT);
			break;
		case MSG_EXTENDED:
			phase = iha_msgin_extended(sc);
			break;
		case MSG_IGN_WIDE_RESIDUE:
			phase = iha_msgin_ignore_wid_resid(sc);
			break;
		case MSG_CMDCOMPLETE:
			sc->sc_flags |= FLAG_EXPECT_DONE_DISC;
			iha_write_8(sc, TUL_SCTRL0, RSFIFO);
			phase = iha_wait(sc, MSG_ACCEPT);
			if (phase != -1) {
				iha_bad_seq(sc);
				return (-1);
			}
			break;
		default:
			printf("[debug] iha_msgin: bad msg type: %d\n", msg);
			phase = iha_msgout_reject(sc);
			break;
		}

		if (phase != PHASE_MSG_IN)
			return (phase);
	}
	/* NOTREACHED */
}

static int
iha_msgin_ignore_wid_resid(struct iha_softc *sc)
{
	int phase;

	phase = iha_wait(sc, MSG_ACCEPT);

	if (phase == PHASE_MSG_IN) {
		phase = iha_wait(sc, XF_FIFO_IN);

		if (phase != -1) {
			iha_write_8(sc, TUL_SFIFO, 0);
			iha_read_8 (sc, TUL_SFIFO);
			iha_read_8 (sc, TUL_SFIFO);

			phase = iha_wait(sc, MSG_ACCEPT);
		}
	}

	return (phase);
}

static int
iha_msgin_extended(struct iha_softc *sc)
{
	uint16_t flags;
	int i, phase, msglen, msgcode;

	/* XXX - can we just stop reading and reject, or do we have to
	 *	 read all input, discarding the excess, and then reject
	 */
	for (i = 0; i < IHA_MAX_EXTENDED_MSG; i++) {
		phase = iha_wait(sc, MSG_ACCEPT);

		if (phase != PHASE_MSG_IN)
			return (phase);

		iha_write_32(sc, TUL_STCNT0, 1);

		if (iha_wait(sc, XF_FIFO_IN) == -1)
			return (-1);

		sc->sc_msg[i] = iha_read_8(sc, TUL_SFIFO);

		if (sc->sc_msg[0] == i)
			break;
	}

	msglen	= sc->sc_msg[0];
	msgcode = sc->sc_msg[1];

	if ((msglen == MSG_EXT_SDTR_LEN) && (msgcode == MSG_EXT_SDTR)) {
		if (iha_msgin_sdtr(sc) == 0) {
			iha_sync_done(sc);
			return (iha_wait(sc, MSG_ACCEPT));
		}

		iha_set_ssig(sc, REQ | BSY | SEL, ATN);

		phase = iha_wait(sc, MSG_ACCEPT);
		if (phase != PHASE_MSG_OUT)
			return (phase);

		/* Clear FIFO for important message - final SYNC offer */
		iha_write_8(sc, TUL_SCTRL0, RSFIFO);

		iha_sync_done(sc);		/* This is our final offer */

	} else if ((msglen == MSG_EXT_WDTR_LEN) && (msgcode == MSG_EXT_WDTR)) {

		flags = sc->sc_actscb->scb_tcs->tcs_flags;

		if ((flags & FLAG_NO_WIDE) != 0)
			/* Offer 8bit xfers only */
			sc->sc_msg[2] = MSG_EXT_WDTR_BUS_8_BIT;

		else if (sc->sc_msg[2] > MSG_EXT_WDTR_BUS_32_BIT)
			/* BAD MSG */
			return (iha_msgout_reject(sc));

		else if (sc->sc_msg[2] == MSG_EXT_WDTR_BUS_32_BIT)
			/* Offer 16bit instead */
			sc->sc_msg[2] = MSG_EXT_WDTR_BUS_16_BIT;

		else {
			iha_wide_done(sc);
			if ((flags & FLAG_NO_NEG_SYNC) == 0)
				iha_set_ssig(sc, REQ | BSY | SEL, ATN);
			return (iha_wait(sc, MSG_ACCEPT));
		}

		iha_set_ssig(sc, REQ | BSY | SEL, ATN);

		phase = iha_wait(sc, MSG_ACCEPT);
		if (phase != PHASE_MSG_OUT)
			return (phase);

	} else
		return (iha_msgout_reject(sc));

	/* Send message built in sc->sc_msg[] */
	return (iha_msgout_extended(sc));
}

/*
 * iha_msgin_sdtr - check SDTR msg in sc_msg. If the offer is
 *                  acceptable leave sc_msg as is and return 0.
 *                  If the negotiation must continue, modify sc_msg
 *                  as needed and return 1. Else return 0.
 */
static int
iha_msgin_sdtr(struct iha_softc *sc)
{
	uint16_t flags;
	uint8_t default_period;
	int newoffer;

	flags = sc->sc_actscb->scb_tcs->tcs_flags;

	default_period = iha_rate_tbl[flags & FLAG_SCSI_RATE];

	if (sc->sc_msg[3] == 0) /* target offered async only. Accept it. */
		return (0);

	newoffer = 0;

	if ((flags & FLAG_NO_SYNC) != 0) {
		sc->sc_msg[3] = 0;
		newoffer      = 1;
	}

	if (sc->sc_msg[3] > IHA_MAX_TARGETS - 1) {
		sc->sc_msg[3] = IHA_MAX_TARGETS - 1;
		newoffer      = 1;
	}

	if (sc->sc_msg[2] < default_period) {
		sc->sc_msg[2] = default_period;
		newoffer      = 1;
	}

	if (sc->sc_msg[2] > IHA_MAX_PERIOD) {
		sc->sc_msg[3] = 0;
		newoffer      = 1;
	}

	return (newoffer);
}

static int
iha_msgout(struct iha_softc *sc, uint8_t msg)
{
	iha_write_8(sc, TUL_SFIFO, msg);

	return (iha_wait(sc, XF_FIFO_OUT));
}

static void
iha_msgout_abort(struct iha_softc *sc, uint8_t aborttype)
{
	iha_set_ssig(sc, REQ | BSY | SEL, ATN);

	switch (iha_wait(sc, MSG_ACCEPT)) {
	case -1:
		break;

	case PHASE_MSG_OUT:
		sc->sc_flags |= FLAG_EXPECT_DISC;
		if (iha_msgout(sc, aborttype) != -1)
			iha_bad_seq(sc);
		break;

	default:
		iha_bad_seq(sc);
		break;
	}
}

static int
iha_msgout_reject(struct iha_softc *sc)
{
	iha_set_ssig(sc, REQ | BSY | SEL, ATN);

	if (iha_wait(sc, MSG_ACCEPT) == PHASE_MSG_OUT)
		return (iha_msgout(sc, MSG_MESSAGE_REJECT));

	return (-1);
}

static int
iha_msgout_extended(struct iha_softc *sc)
{
	int phase;

	iha_write_8(sc, TUL_SFIFO, MSG_EXTENDED);

	bus_space_write_multi_1(sc->sc_iot, sc->sc_ioh, TUL_SFIFO,
	    sc->sc_msg, sc->sc_msg[0] + 1);

	phase = iha_wait(sc, XF_FIFO_OUT);

	iha_write_8(sc, TUL_SCTRL0, RSFIFO);
	iha_set_ssig(sc, REQ | BSY | SEL | ATN, 0);

	return (phase);
}

static int
iha_msgout_wdtr(struct iha_softc *sc)
{
	sc->sc_actscb->scb_tcs->tcs_flags |= FLAG_WIDE_DONE;

	sc->sc_msg[0] = MSG_EXT_WDTR_LEN;
	sc->sc_msg[1] = MSG_EXT_WDTR;
	sc->sc_msg[2] = MSG_EXT_WDTR_BUS_16_BIT;

	return (iha_msgout_extended(sc));
}

static int
iha_msgout_sdtr(struct iha_softc *sc)
{
	uint16_t rateindex;
	uint8_t sync_rate;

	rateindex = sc->sc_actscb->scb_tcs->tcs_flags & FLAG_SCSI_RATE;

	sync_rate = iha_rate_tbl[rateindex];

	/* XXX: NetBSD does: tcs->flags |= FLAG_SYNC_DONE; */

	sc->sc_msg[0] = MSG_EXT_SDTR_LEN;
	sc->sc_msg[1] = MSG_EXT_SDTR;
	sc->sc_msg[2] = sync_rate;
	sc->sc_msg[3] = IHA_MAX_TARGETS - 1; /* REQ/ACK */

	return (iha_msgout_extended(sc));
}

static void
iha_wide_done(struct iha_softc *sc)
{
	struct tcs *tcs = sc->sc_actscb->scb_tcs;

	tcs->tcs_js_period = 0;

	if (sc->sc_msg[2] != 0)
		tcs->tcs_js_period |= PERIOD_WIDE_SCSI;

	tcs->tcs_sconfig0 &= ~ALTPD;
	tcs->tcs_flags	  &= ~FLAG_SYNC_DONE;
	tcs->tcs_flags	  |=  FLAG_WIDE_DONE;

	iha_write_8(sc, TUL_SCONFIG0, tcs->tcs_sconfig0);
	iha_write_8(sc, TUL_SYNCM,    tcs->tcs_js_period);
}

static void
iha_sync_done(struct iha_softc *sc)
{
	struct tcs *tcs = sc->sc_actscb->scb_tcs;
	int i;

	if ((tcs->tcs_flags & FLAG_SYNC_DONE) == 0) {
		if (sc->sc_msg[3] != 0) {
			tcs->tcs_js_period |= sc->sc_msg[3];

			/* pick the highest possible rate */
			for (i = 0; i < sizeof(iha_rate_tbl); i++)
				if (iha_rate_tbl[i] >= sc->sc_msg[2])
					break;

			tcs->tcs_js_period |= (i << 4);
			tcs->tcs_sconfig0  |= ALTPD;
		}

		tcs->tcs_flags |= FLAG_SYNC_DONE;

		iha_write_8(sc, TUL_SCONFIG0, tcs->tcs_sconfig0);
		iha_write_8(sc, TUL_SYNCM,    tcs->tcs_js_period);
	}
}

static void
iha_reset_chip(struct iha_softc *sc)
{
	/* reset tulip chip */

	iha_write_8(sc, TUL_SCTRL0, RSCSI);

	do
		sc->sc_jsint = iha_read_8(sc, TUL_SISTAT);
	while ((sc->sc_jsint & SRSTD) == 0);

	iha_set_ssig(sc, 0, 0);

	iha_read_8(sc, TUL_SISTAT);	/* Clear any active interrupt */
}

static void
iha_select(struct iha_softc *sc, struct iha_scb *scb, uint8_t select_type)
{
	switch (select_type) {
	case SEL_ATN:
		iha_write_8(sc, TUL_SFIFO, scb->scb_ident);
		bus_space_write_multi_1(sc->sc_iot, sc->sc_ioh, TUL_SFIFO,
		    scb->scb_cdb, scb->scb_cdb_len);

		scb->scb_next_stat = 2;
		break;

	case SELATNSTOP:
		scb->scb_next_stat = 1;
		break;

	case SEL_ATN3:
		iha_write_8(sc, TUL_SFIFO, scb->scb_ident);
		iha_write_8(sc, TUL_SFIFO, scb->scb_tag_msg);
		iha_write_8(sc, TUL_SFIFO, scb->scb_tag_id);

		bus_space_write_multi_1(sc->sc_iot, sc->sc_ioh, TUL_SFIFO,
		    scb->scb_cdb, scb->scb_cdb_len);

		scb->scb_next_stat = 2;
		break;

	default:
		printf("[debug] iha_select() - unknown select type = 0x%02x\n",
		    select_type);
		return;
	}

	iha_del_pend_scb(sc, scb);
	scb->scb_status = STATUS_SELECT;

	sc->sc_actscb = scb;

	iha_write_8(sc, TUL_SCMD, select_type);
}

/*
 * iha_wait - wait for an interrupt to service or a SCSI bus phase change
 *            after writing the supplied command to the tulip chip. If
 *            the command is NO_OP, skip the command writing.
 */
static int
iha_wait(struct iha_softc *sc, uint8_t cmd)
{
	if (cmd != NO_OP)
		iha_write_8(sc, TUL_SCMD, cmd);

	/*
	 * Have to do this here, in addition to in iha_isr, because
	 * interrupts might be turned off when we get here.
	 */
	do
		sc->sc_jsstatus0 = iha_read_8(sc, TUL_STAT0);
	while ((sc->sc_jsstatus0 & INTPD) == 0);

	sc->sc_jsstatus1 = iha_read_8(sc, TUL_STAT1);
	sc->sc_jsint     = iha_read_8(sc, TUL_SISTAT);

	sc->sc_phase = sc->sc_jsstatus0 & PH_MASK;

	if ((sc->sc_jsint & SRSTD) != 0) {
		/* SCSI bus reset interrupt */
		iha_reset_scsi_bus(sc);
		return (-1);
	}

	if ((sc->sc_jsint & RSELED) != 0)
		/* Reselection interrupt */
		return (iha_resel(sc));

	if ((sc->sc_jsint & STIMEO) != 0) {
		/* selected/reselected timeout interrupt */
		iha_busfree(sc);
		return (-1);
	}

	if ((sc->sc_jsint & DISCD) != 0) {
		/* BUS disconnection interrupt */
		if ((sc->sc_flags & FLAG_EXPECT_DONE_DISC) != 0) {
			iha_write_8(sc, TUL_SCTRL0,   RSFIFO);
			iha_write_8(sc, TUL_SCONFIG0, SCONFIG0DEFAULT);
			iha_write_8(sc, TUL_SCTRL1,   EHRSL);
			iha_append_done_scb(sc, sc->sc_actscb, HOST_OK);
			sc->sc_flags &= ~FLAG_EXPECT_DONE_DISC;

		} else if ((sc->sc_flags & FLAG_EXPECT_DISC) != 0) {
			iha_write_8(sc, TUL_SCTRL0,   RSFIFO);
			iha_write_8(sc, TUL_SCONFIG0, SCONFIG0DEFAULT);
			iha_write_8(sc, TUL_SCTRL1,   EHRSL);
			sc->sc_actscb = NULL;
			sc->sc_flags &= ~FLAG_EXPECT_DISC;

		} else
			iha_busfree(sc);

		return (-1);
	}

	return (sc->sc_phase);
}

/*
 * iha_done_scb - We have a scb which has been processed by the
 *                adaptor, now we look to see how the operation went.
 */
static void
iha_done_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	union ccb		*ccb;

	ccb = scb->scb_ccb;
	if (ccb != NULL) {
		ccb->csio.scsi_status = scb->scb_ta_stat;

		switch (scb->scb_ha_stat) {
		case HOST_OK:
			switch (scb->scb_ta_stat) {
			case SCSI_STATUS_OK:
				ccb->ccb_h.status |= CAM_REQ_CMP;
				break;

			case SCSI_STATUS_CMD_TERMINATED:
			case SCSI_STATUS_CHECK_COND:
				ccb->ccb_h.status |= CAM_AUTOSNS_VALID;
				ccb->csio.sense_data = scb->scb_scsi_sense_data;
				break;

			default:
				ccb->ccb_h.status |= CAM_SCSI_STATUS_ERROR;
				break;
			}
			break;

		case HOST_SEL_TOUT:
			ccb->ccb_h.status = CAM_SEL_TIMEOUT;
			break;

		case HOST_SCSI_RST:
		case HOST_DEV_RST:
			ccb->ccb_h.status = CAM_SCSI_BUS_RESET;
			break;

		case HOST_SPERR:
			device_printf(sc->sc_dev, "SCSI Parity error "
			    "detected\n");
			ccb->ccb_h.status = CAM_UNCOR_PARITY;
			break;

		case HOST_TIMED_OUT:
			ccb->ccb_h.status = CAM_CMD_TIMEOUT;
			break;

		case HOST_DO_DU:
			ccb->ccb_h.status = CAM_DATA_RUN_ERR;
			break;

		default:
			ccb->ccb_h.status = CAM_UNREC_HBA_ERROR;
			break;
		}

		if ((ccb->ccb_h.flags & CAM_DIR_MASK) != CAM_DIR_NONE) {
			bus_dmasync_op_t	op;

			if ((ccb->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN)
				op = BUS_DMASYNC_POSTREAD;
			else
				op = BUS_DMASYNC_POSTWRITE;

			bus_dmamap_sync(sc->sc_buf_dmat, scb->scb_dmamap, op);
			bus_dmamap_unload(sc->sc_buf_dmat, scb->scb_dmamap);
		}

		untimeout(iha_timeout, scb, scb->scb_ccb->ccb_h.timeout_ch);
		ccb->ccb_h.status |= CAM_REQ_CMP;
		xpt_done(ccb);
	}

	iha_append_free_scb(sc, scb);
}

static void
iha_timeout(void *arg)
{
	struct iha_scb		*scb;
	union ccb		*ccb;
	struct iha_softc	*sc;

	scb = arg;
	ccb = scb->scb_ccb;

	if (ccb == NULL)
		printf("[debug] iha_timeout called with ccb == NULL\n");

	else {
		sc = ccb->ccb_h.ccb_iha_ptr;
		xpt_print_path(ccb->ccb_h.path);
		printf("SCSI OpCode 0x%02x timed out\n", ccb->ccb_h.func_code);

		iha_abort_ccb(sc, ccb, HOST_TIMED_OUT);
	}
}

static void
iha_start_scb(struct iha_softc *sc, struct iha_scb *scb)
{
	int s;

	IHA_TRACE;
	IHA_PRINTF(("iha_start_scb: scb_flags = %x\n",
	    (unsigned)scb->scb_flags));

	s = splbio();

	if (scb->scb_cdb[0] == REQUEST_SENSE)
		iha_push_pend_scb(sc, scb);   /* Insert SCB at head of Pend */
	else
		iha_append_pend_scb(sc, scb); /* Append SCB to tail of Pend */

	/*
	 * Run through iha_main() to ensure something is active, if
	 * only this new SCB.
	 */
	if (sc->sc_semaph != SEMAPH_IN_MAIN) {
		iha_disable_ints(sc);
		sc->sc_semaph = SEMAPH_IN_MAIN;;

		splx(s);
		iha_main(sc);
		s = splbio();

		sc->sc_semaph = ~SEMAPH_IN_MAIN;;
		iha_enable_ints(sc);
	}

	splx(s);
}


/*
 * iha_set_ssig - read the current scsi signal mask, then write a new
 *		  one which turns off/on the specified signals.
 */
static void
iha_set_ssig(struct iha_softc *sc, uint8_t offsigs, uint8_t onsigs)
{
	uint8_t currsigs;

	currsigs = iha_read_8(sc, TUL_SSIGI);
	iha_write_8(sc, TUL_SSIGO, (currsigs & ~offsigs) | onsigs);
}

static void
iha_exec_scb(void *arg, bus_dma_segment_t *psegs, int nsegs, int error)
{
	struct iha_scb		*scb;
	union ccb		*ccb;
	struct iha_softc	*sc;
	int			i;
	struct iha_sg_element	*sg;
	bus_dmasync_op_t	op;

	scb = arg;
	ccb = scb->scb_ccb;
	sc  = ccb->ccb_h.ccb_iha_ptr;

	IHA_TRACE;

	if (error != 0) {
		if (error != EFBIG)
			device_printf(sc->sc_dev, "iha_exec_scb: error = "
			    "%d\n", error);
		else  {
                        xpt_freeze_devq(ccb->ccb_h.path, 1);
                        ccb->ccb_h.status = CAM_REQ_TOO_BIG | CAM_DEV_QFRZN;
		}
		xpt_done(ccb);
		IHA_TRACE;

		iha_append_free_scb(sc, scb);
		return;
	}

	scb->scb_buf_len = ccb->csio.dxfer_len;

	if (nsegs > 1) {
		sg = scb->scb_sg_list;

		for (i = 0; i < nsegs; i++) {
			/* XXX: NetBSD uses htole32? */
			sg[i].sg_ptr = psegs[i].ds_addr;
			sg[i].sg_len = psegs[i].ds_len;
		}

		scb->scb_flags |= CAM_SCATTER_VALID;
		scb->scb_sg_len = nsegs;
		scb->scb_sg_max = nsegs;

		scb->scb_buf_paddr = scb->scb_sg_paddr;
	} else
		scb->scb_buf_paddr = psegs[0].ds_addr;

	if ((scb->scb_flags & CAM_DIR_MASK) == CAM_DIR_IN)
		op = BUS_DMASYNC_PREREAD;
	else
		op = BUS_DMASYNC_PREWRITE;
	bus_dmamap_sync(sc->sc_buf_dmat, scb->scb_dmamap, op);

	ccb->ccb_h.timeout_ch = timeout(iha_timeout, scb,
	    (ccb->ccb_h.timeout * hz) / 1000);

	iha_start_scb(sc, scb);
}

/*
 * iha_alloc_scbs - allocate and map the SCBs for the supplied iha_softc
 */
static int
iha_alloc_scbs(struct iha_softc *sc)
{
	int		error;
	bus_size_t	memsize;
	void		*p;

	memsize = sizeof(struct iha_scb) * IHA_MAX_SCB;

	/* Create a DMA tag for buffers */
	if ((error = bus_dma_tag_create(NULL, 1, 0, BUS_SPACE_MAXADDR_32BIT,
	    BUS_SPACE_MAXADDR, NULL, NULL, MAXBSIZE, IHA_MAX_SG_ENTRIES,
	    BUS_SPACE_MAXSIZE_32BIT, BUS_DMA_ALLOCNOW, &sc->sc_buf_dmat))
	    != 0) {
		device_printf(sc->sc_dev, "cannote create DMA tag\n");
		return (error);
	}

	/* Create a DMA tag the SCBs */
	if ((error = bus_dma_tag_create(NULL, 1, 0, BUS_SPACE_MAXADDR_32BIT,
	    BUS_SPACE_MAXADDR, NULL, NULL, memsize, 1,
	    BUS_SPACE_MAXSIZE_32BIT, 0, &sc->sc_scb_dmat)) != 0) {
		device_printf(sc->sc_dev, "cannote create DMA tag\n");
		goto failed1;
	}

	/* Allocate dma-safe memory for the SCBs */
	if ((error = bus_dmamem_alloc(sc->sc_scb_dmat, &p, 0, &sc->sc_dmamap))
	    != 0) {
		device_printf(sc->sc_dev, "cannot allocate SCBs\n");
		goto failed2;
	}
	sc->sc_scb = p;

	/* Create and load the DMA map used for the SCBs */
	if ((error = bus_dmamap_load(sc->sc_scb_dmat, sc->sc_dmamap,
	    sc->sc_scb, memsize, iha_dmamap_helper, sc, 0)) != 0) {
		device_printf(sc->sc_dev, "unable to load control DMA map, "
		    "error = %d\n", error);
		goto failed3;
	}
	bzero(sc->sc_scb, sizeof(struct iha_scb) * IHA_MAX_SCB);

	sc->sc_alloc_flags |= IHA_ALLOC_SCBS;

	return (0);

failed3:
	bus_dmamem_free(sc->sc_scb_dmat, sc->sc_scb, sc->sc_dmamap);
failed2:
	bus_dma_tag_destroy(sc->sc_scb_dmat);
failed1:
	bus_dma_tag_destroy(sc->sc_buf_dmat);

	return (error);
}

/*
 * iha_read_eeprom - read contents of serial EEPROM into iha_nvram pointed at
 *                                        by parameter nvram.
 */
static void
iha_read_eeprom(struct iha_softc *sc, struct iha_nvram *nvram)
{
	uint32_t chksum;
	uint16_t *np;
	uint8_t gctrl, addr;

	const int chksum_addr = offsetof(struct iha_nvram, nvm_check_sum) / 2;

	/* Enable EEProm programming */
	gctrl = iha_read_8(sc, TUL_GCTRL0) | EEPRG;
	iha_write_8(sc, TUL_GCTRL0, gctrl);

	/* Read EEProm */
	np = (uint16_t *)nvram;
	for (addr = 0, chksum = 0; addr < chksum_addr; addr++, np++) {
		*np = iha_se2_rd(sc, addr);
		chksum += *np;
	}

	chksum &= 0x0000ffff;
	nvram->nvm_check_sum = iha_se2_rd(sc, chksum_addr);

	/* Disable EEProm programming */
	gctrl = iha_read_8(sc, TUL_GCTRL0) & ~EEPRG;
	iha_write_8(sc, TUL_GCTRL0, gctrl);

	if ((nvram->nvm_signature != SIGNATURE) ||
	    (nvram->nvm_check_sum != chksum))
		panic("iha: invalid EEPROM,  bad signature or checksum");
}

/*
 * iha_se2_rd - read & return the 16 bit value at the specified
 *		offset in the Serial E2PROM
 *
 */
static uint16_t
iha_se2_rd(struct iha_softc *sc, uint8_t addr)
{
	uint16_t readWord;
	uint8_t bit;
	int i;

	/* Send 'READ' instruction == address | READ bit */
	iha_se2_instr(sc, addr | READ);

	readWord = 0;
	for (i = 15; i >= 0; i--) {
		iha_write_8(sc, TUL_NVRAM, NVRCS | NVRCK);
		DELAY(5);

		iha_write_8(sc, TUL_NVRAM, NVRCS);
		DELAY(5);

		/* sample data after the following edge of clock     */
		bit = iha_read_8(sc, TUL_NVRAM) & NVRDI;
		DELAY(5);

		readWord += bit << i;
	}

	iha_write_8(sc, TUL_NVRAM, 0);
	DELAY(5);

	return (readWord);
}

/*
 * iha_se2_instr - write an octet to serial E2PROM one bit at a time
 */
static void
iha_se2_instr(struct iha_softc *sc, uint8_t instr)
{
	uint8_t b;
	int i;

	b = NVRCS | NVRDO; /* Write the start bit (== 1) */

	iha_write_8(sc, TUL_NVRAM, b);
	DELAY(5);
	iha_write_8(sc, TUL_NVRAM, b | NVRCK);
	DELAY(5);

	for (i = 0; i < 8; i++, instr <<= 1) {
		if (instr & 0x80)
			b = NVRCS | NVRDO; /* Write a 1 bit */
		else
			b = NVRCS;	   /* Write a 0 bit */

		iha_write_8(sc, TUL_NVRAM, b);
		DELAY(5);
		iha_write_8(sc, TUL_NVRAM, b | NVRCK);
		DELAY(5);
	}

	iha_write_8(sc, TUL_NVRAM, NVRCS);
	DELAY(5);

	return;
}

/*
 * iha_reset_tcs - reset the target control structure pointed
 *		   to by tcs to default values. tcs_flags
 *		   only has the negotiation done bits reset as
 *		   the other bits are fixed at initialization.
 */
static void
iha_reset_tcs(struct tcs *tcs, uint8_t config0)
{
	tcs->tcs_flags	   &= ~(FLAG_SYNC_DONE | FLAG_WIDE_DONE);
	tcs->tcs_js_period  = 0;
	tcs->tcs_sconfig0   = config0;
	tcs->tcs_tag_cnt    = 0;
	tcs->tcs_nontag_scb = NULL;
}

/*
 * iha_scsi_io - handle XPT_SCSI_IO for iha_action()
 */
static void
iha_scsi_io(struct iha_softc *sc, union ccb *ccb)
{
	struct ccb_hdr		*ccb_h;
	struct ccb_scsiio	*csio;
	struct iha_scb		*scb;
	int			error;
	int			flags;

	IHA_TRACE;

	ccb_h = &ccb->ccb_h;
	flags = ccb_h->flags;
	csio  = &ccb->csio;

	IHA_PRINTF(("iha_scsi_io: flags = %x status = %x\n", ccb_h->flags,
	    ccb_h->status));

	if ((ccb_h->status & CAM_STATUS_MASK) != CAM_REQ_INPROG) {
		xpt_done(ccb);
		return;
	}
	if (csio->cdb_len > 12) {
		ccb_h->status = CAM_REQ_INVALID;
		xpt_done(ccb);
		return;
	}
	if (ccb_h->target_id > IHA_MAX_TARGETS ||
	    ccb_h->target_id == sc->sc_adapter_target) {
		ccb_h->status = CAM_TID_INVALID;
		xpt_done(ccb);
		return;
	}

	if ((scb = iha_pop_free_scb(sc)) == NULL) {
		ccb->ccb_h.status = CAM_RESRC_UNAVAIL;
		xpt_done(ccb);
		return;
	}

	scb->scb_target  = ccb_h->target_id;
	scb->scb_lun     = ccb_h->target_lun;
	scb->scb_tcs     = &sc->sc_tcs[scb->scb_target];
	scb->scb_flags   = flags;
	scb->scb_cdb_len = csio->cdb_len;
	scb->scb_ccb     = ccb;
	scb->scb_ident   = MSG_IDENTIFYFLAG |
		(scb->scb_lun & MSG_IDENTIFY_LUNMASK);

	ccb->ccb_h.ccb_iha_ptr = sc;

	if ((flags & CAM_CDB_POINTER) != 0) {
		if ((flags & CAM_CDB_PHYS) != 0) {
			ccb->ccb_h.status = CAM_REQ_INVALID;
			xpt_done(ccb);
			iha_append_free_scb(sc, scb);
			return;
		} else
			bcopy(csio->cdb_io.cdb_ptr, scb->scb_cdb,
			    scb->scb_cdb_len);
	} else
		bcopy(csio->cdb_io.cdb_bytes, scb->scb_cdb, scb->scb_cdb_len);

	if ((flags & CAM_DIR_MASK) != CAM_DIR_NONE) {
		if ((flags & CAM_SCATTER_VALID) == 0) {
			/* We've been given a pointer to a single
			 * buffer. */
			if ((flags & CAM_DATA_PHYS) == 0) {
				error = bus_dmamap_load(sc->sc_buf_dmat,
				    scb->scb_dmamap, csio->data_ptr,
				    csio->dxfer_len, iha_exec_scb, scb, 0);
				if (error == EINPROGRESS) {
					/* So as to maintain ordering,
					 * freeze the controller queue
					 * until our mapping is
					 * returned. */
					xpt_freeze_simq(sc->sc_sim, 1);
					csio->ccb_h.status |= CAM_RELEASE_SIMQ;
				}
			} else {
				/* Pointer to physical buffer */
				struct bus_dma_segment	seg;

				seg.ds_addr = (bus_addr_t)csio->data_ptr;
				seg.ds_len  = csio->dxfer_len;
				iha_exec_scb(scb, &seg, 1, 0);
			}
		} else {
			struct bus_dma_segment	*pseg;

			if ((flags & CAM_DATA_PHYS) != 0 ||
			    (flags & CAM_SG_LIST_PHYS) == 0) {
				ccb->ccb_h.status = CAM_REQ_INVALID;
				xpt_done(ccb);
				iha_append_free_scb(sc, scb);
				return;
			}

			pseg = (struct bus_dma_segment *)csio->data_ptr;
			iha_exec_scb(scb, pseg, csio->sglist_cnt, 0);
		}
	} else {
		ccb->ccb_h.timeout_ch = timeout(iha_timeout, scb,
		    (ccb->ccb_h.timeout * hz) / 1000);
		iha_start_scb(sc, scb);
	}
}

/*
 * iha_action - CAM callback to start executing the CCB
 */
static void
iha_action(struct cam_sim *sim, union ccb *ccb)
{
	struct iha_softc		*sc;
	struct ccb_trans_settings	*cts;
	struct ccb_calc_geometry	*ccg;
	struct ccb_pathinq		*cpi;
	uint32_t			size_mb;
	uint32_t			secs_per_cylinder;

	sc = cam_sim_softc(sim);

	switch (ccb->ccb_h.func_code) {
	case XPT_SCSI_IO:
		iha_scsi_io(sc, ccb);
		break;

	case XPT_GET_TRAN_SETTINGS:
		IHA_TRACE;

		/* Get default/user set transfer settings for the target */
		cts = &ccb->cts;

		cts->flags = CCB_TRANS_DISC_ENB | CCB_TRANS_TAG_ENB;

		cts->bus_width = (sc->sc_target_max == 16) ?
			MSG_EXT_WDTR_BUS_16_BIT : MSG_EXT_WDTR_BUS_8_BIT;

		cts->sync_period = 12;
		cts->sync_offset = 15;

		cts->valid = CCB_TRANS_SYNC_RATE_VALID
			   | CCB_TRANS_SYNC_OFFSET_VALID
			   | CCB_TRANS_BUS_WIDTH_VALID
			   | CCB_TRANS_DISC_VALID
			   | CCB_TRANS_TQ_VALID;

		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;

	case XPT_SET_TRAN_SETTINGS:
		IHA_TRACE;

		cts = &ccb->cts;

		IHA_PRINTF(("set_trans_settings: valid = %x\n", cts->flags)); 

		/* XXX: implement */

		ccb->ccb_h.status = CAM_PROVIDE_FAIL;
		xpt_done(ccb);
		break;

	case XPT_CALC_GEOMETRY:
		IHA_TRACE;

		ccg = &ccb->ccg;
		size_mb = ccg->volume_size / (1024L * 1024L / ccg->block_size);

		if (size_mb >= 1024 ) {
			ccg->heads = 255;
			ccg->secs_per_track = 63;
		} else {
			ccg->heads = 64;
			ccg->secs_per_track = 32;
		}
		secs_per_cylinder = ccg->heads * ccg->secs_per_track;
		ccg->cylinders = ccg->volume_size / secs_per_cylinder;
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;

	case XPT_RESET_BUS:
		IHA_TRACE;

		iha_reset_chip(sc);
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;

	case XPT_PATH_INQ:
		IHA_TRACE;

		cpi = &ccb->cpi;

		cpi->version_num = 1;
		cpi->hba_inquiry = PI_SDTR_ABLE;
		if (ccb->csio.tag_action != 0)
			cpi->hba_inquiry |= PI_TAG_ABLE;
		if (sc->sc_target_max == 16)
			cpi->hba_inquiry |= PI_WIDE_16;

		cpi->target_sprt = 0;
		cpi->hba_misc = 0;
		cpi->hba_eng_cnt = 0;
		cpi->max_target = sc->sc_target_max - 1;
		cpi->max_lun = 7;
		cpi->initiator_id = sc->sc_adapter_target;
		cpi->bus_id = cam_sim_bus(sim);
		strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
		strncpy(cpi->hba_vid, "Initio", HBA_IDLEN);
		strncpy(cpi->dev_name, cam_sim_name(sim), DEV_IDLEN);
		cpi->unit_number = cam_sim_unit(sim);
		cpi->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;

	default:
		IHA_PRINTF(("iha %s:%d func_code = %d\n", __func__, __LINE__,
		    ccb->ccb_h.func_code));

		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;
	}
}

/*
 * iha_poll - CAM callback
 */
static void
iha_poll(struct cam_sim *sim)
{
	struct iha_softc	*sc;

	IHA_TRACE;

	sc = cam_sim_softc(sim);

	iha_intr(sc);
}

/*
 * iha_chipname - return the chipname as string.
 */
static const char *
iha_chipname(device_t dev)
{
	switch (pci_get_device(dev)) {
	case PCI_PRODUCT_INITIO_INIC940:
		return ("Initio INIC-940 PCI SCSI adapter");
	case PCI_PRODUCT_INITIO_INIC941:
		return ("Initio INIC-941 PCI SCSI adapter");
	case PCI_PRODUCT_INITIO_INIC950:
		return ("Initio INI-9XXXU/UW PCI SCSI adapter");
	}
	return (NULL);
}

/*
 * iha_probe - check whether the device is supported by this driver
 */
static int
iha_probe(device_t dev)
{
	const char	*name;

	if (pci_get_vendor(dev) != PCI_VENDOR_INITIO)
		return (ENXIO);

	if ((name = iha_chipname(dev)) == NULL)
		return (ENXIO);

	device_set_desc(dev, name);
	return (0);
}


/*
 * iha_dmamap_helper - callback for bus_dmamap_load()
 */
static void
iha_dmamap_helper(void *arg, bus_dma_segment_t *segs, int nsegs, int error)
{
	struct iha_softc	*sc;

	sc = arg;
	sc->sc_scb_phys = segs->ds_addr;
}

/*
 * iha_cam_attach - register this driver at the CAM subsystem
 */
static int
iha_cam_attach(struct iha_softc *sc)
{
	int			openings;
	struct cam_devq		*devq;

	/*
	 * Create the device queue.  We only allow MAX_START-1 concurrent
	 * transactions so we can be sure to have one element free in our
	 * start queue to reset to the idle loop.
	 */
	openings = IHA_MAX_SCB - 1;
	if ((devq = cam_simq_alloc(openings)) == NULL){
		device_printf(sc->sc_dev, "cam_simq_alloc() failed!\n");
		return (ENOMEM);
	}

	/* Now tell the generic SCSI layer about our bus. */
	if ((sc->sc_sim = cam_sim_alloc(iha_action, iha_poll, "iha", sc,
	    sc->sc_unit, 1, openings, devq)) == NULL) {
		device_printf(sc->sc_dev, "cam_sim_alloc() failed!\n");
		cam_simq_free(devq);
		return (ENOMEM);
	}

	if (xpt_bus_register(sc->sc_sim, 0) != CAM_SUCCESS) {
		device_printf(sc->sc_dev, "xpt_bus_register() failed!\n");
		cam_sim_free(sc->sc_sim, TRUE);
		return (ENOMEM);
	}

	if (xpt_create_path(&sc->sc_path, NULL, cam_sim_path(sc->sc_sim),
	    CAM_TARGET_WILDCARD, CAM_LUN_WILDCARD) != CAM_REQ_CMP) {
		device_printf(sc->sc_dev, "xpt_create_path() failed!\n");
		xpt_bus_deregister(cam_sim_path(sc->sc_sim));
		cam_sim_free(sc->sc_sim, TRUE);
		return (ENOMEM);
	}

	return (0);
}

/*
 * iha_cam_release - release all CAM resources
 */
static void
iha_cam_release(struct iha_softc *sc)
{
	if (sc->sc_sim != NULL) {
		xpt_free_path(sc->sc_path);
		xpt_bus_deregister(cam_sim_path(sc->sc_sim));
		cam_sim_free(sc->sc_sim, TRUE);
		sc->sc_sim = NULL;
	}
}

/*
 * iha_attach - XXX: write me
 */
static int
iha_attach(device_t dev)
{
	struct iha_softc	*sc;
	int			error;

	/* XXX: check error codes */

	if ((sc = device_get_softc(dev)) == NULL) {
		device_printf(dev, "device_get_softc() failed\n");
		error = ENOMEM;
		goto failed;
	}
	bzero(sc, sizeof(*sc));

	sc->sc_dev  = dev;
	sc->sc_unit = device_get_unit(dev);

	sc->sc_ioid = PCIR_MAPS;
	if ((sc->sc_iores = bus_alloc_resource(dev, SYS_RES_IOPORT,
	    &sc->sc_ioid, 0lu, ~0lu, 1, RF_ACTIVE)) == NULL) {
		device_printf(dev, "IO resource allocation failed\n");
		error = ENOMEM;
		goto failed;
	}
	sc->sc_iot = rman_get_bustag(sc->sc_iores);
	sc->sc_ioh = rman_get_bushandle(sc->sc_iores);
	sc->sc_alloc_flags |= IHA_ALLOC_IO;

	if ((error = iha_init_tulip(sc)) != 0) {
		device_printf(dev, "iha_init_tulip failed\n");
		goto failed;
	}

	if ((error = iha_cam_attach(sc)) != 0)
		goto failed;
	sc->sc_alloc_flags |= IHA_ALLOC_CAM;

	if ((sc->sc_irqres = bus_alloc_resource(dev, SYS_RES_IRQ,
	    &sc->sc_irqid, 0lu, ~0lu, 1, RF_ACTIVE | RF_SHAREABLE)) == NULL) {
		device_printf(dev, "IRQ resource allocation failed\n");
		error = ENOMEM;
		goto failed;
	}
	sc->sc_alloc_flags |= IHA_ALLOC_IRQ1;

	if ((error = bus_setup_intr(dev, sc->sc_irqres, INTR_TYPE_CAM |
	    INTR_ENTROPY, iha_intr, sc, &sc->sc_irqcookie)) != 0) {
		device_printf(dev, "bus_setup_intr() failed\n");
		goto failed;
	}
	sc->sc_alloc_flags |= IHA_ALLOC_IRQ2;

	/* Finally turn on interrupts */
	iha_write_8(sc, TUL_SIEN, ALL_INTERRUPTS);

	return (0);

failed:
	iha_release(sc);

	return (error);
}

/*
 * iha_release - free resources
 */
static void
iha_release(struct iha_softc *sc)
{
	if (sc->sc_alloc_flags & IHA_ALLOC_CAM)
		iha_cam_release(sc);
	if (sc->sc_alloc_flags & IHA_ALLOC_SCBS) {
		bus_dmamap_unload(sc->sc_scb_dmat, sc->sc_dmamap);
		bus_dmamem_free(sc->sc_scb_dmat, sc->sc_scb, sc->sc_dmamap);
		bus_dma_tag_destroy(sc->sc_scb_dmat);
		bus_dma_tag_destroy(sc->sc_buf_dmat);
	}
	if (sc->sc_alloc_flags & IHA_ALLOC_IRQ2)
		bus_teardown_intr(sc->sc_dev, sc->sc_irqres, sc->sc_irqcookie);
	if (sc->sc_alloc_flags & IHA_ALLOC_IRQ1)
		bus_release_resource(sc->sc_dev, SYS_RES_IRQ, sc->sc_irqid,
		    sc->sc_irqres);
	if (sc->sc_alloc_flags & IHA_ALLOC_IO)
		bus_release_resource(sc->sc_dev, SYS_RES_IOPORT, sc->sc_ioid,
		    sc->sc_iores);
}

/*
 * iha_detach - free resources
 */
static int
iha_detach(device_t dev)
{
	struct iha_softc	*sc;

	sc = device_get_softc(dev);
	if (sc != NULL) {
		iha_disable_ints(sc);
		iha_reset_scsi_bus(sc);
		iha_release(sc);
	}

	return (0);
}

/* 
 * iha_disable_ints - disable interrupts by masking them all
 */
static void
iha_disable_ints(struct iha_softc *sc)
{
	iha_write_8(sc, TUL_IMSK, MASK_ALL);
}

/*
 * iha_enable_ints - enable the SCSI Complete interrupt
 */
static void
iha_enable_ints(struct iha_softc *sc)
{
	iha_write_8(sc, TUL_IMSK, MASK_ALL & ~MSCMP);
}
